/*
 *  xfmedia - simple gtk2 media player based on xine
 *
 *  Copyright (c) 2004 Brian Tarricone, <bjt23@cornell.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "xfmedia-common.h"
#include "xfmedia-internal.h"
#include "main.h"
#include "mainwin.h"
#include "gtkxine.h"
#include "keybindings.h"

#include <libxfce4util/libxfce4util.h>

typedef void (*XfMediaKeybindFunc)(XfMediaMainwin *mwin, gpointer user_data);
#define KEYBIND_FUNC(name) static void name(XfMediaMainwin *mwin, gpointer user_data)

typedef struct
{
	gchar *keybind_id;
	XfMediaKeybindFunc callback;
	gpointer user_data;
} XfMediaKeybindEntry;

KEYBIND_FUNC(kb_play)
{
	xfmedia_playpause_cb(NULL, mwin);
}

KEYBIND_FUNC(kb_pause)
{
	xfmedia_playpause_cb(NULL, mwin);
}

KEYBIND_FUNC(kb_stop)
{
	xfmedia_stop_cb(NULL, mwin);
}

KEYBIND_FUNC(kb_prev_track)
{
	xfmedia_prev_cb(NULL, mwin);
}

KEYBIND_FUNC(kb_next_track)
{
	xfmedia_next_cb(NULL, mwin);
}

KEYBIND_FUNC(kb_quit)
{
	xfmedia_quit(mwin, XFMEDIA_QUIT_MODE_EXT);
}

KEYBIND_FUNC(kb_fullscreen)
{
	gboolean is_fs = gtk_xine_is_fullscreen(mwin->gtx);
		
	gtk_xine_set_fullscreen(mwin->gtx, !is_fs);
	if(!is_fs)
		gtk_widget_hide(mwin->video_window);
	else
		gtk_widget_show(mwin->video_window);
}

KEYBIND_FUNC(kb_repeat)
{
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mwin->repeat_mi),
			!gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mwin->repeat_mi)));
}

KEYBIND_FUNC(kb_shuffle)
{
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mwin->shuffle_mi),
			!gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mwin->shuffle_mi)));
}

KEYBIND_FUNC(kb_audio_vis)
{
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mwin->svis_mi),
			!gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mwin->svis_mi)));
}

KEYBIND_FUNC(kb_faster)
{
	gint cur_speed = gtk_xine_get_param(mwin->gtx, XINE_PARAM_SPEED);
	gint new_speed;
	
	switch(cur_speed) {
		case XINE_SPEED_PAUSE:
			new_speed = XINE_SPEED_SLOW_4;
			break;
		case XINE_SPEED_SLOW_4:
			new_speed = XINE_SPEED_SLOW_2;
			break;
		case XINE_SPEED_SLOW_2:
			new_speed = XINE_SPEED_NORMAL;
			break;
		case XINE_SPEED_NORMAL:
			new_speed = XINE_SPEED_FAST_2;
			break;
		case XINE_SPEED_FAST_2:
			new_speed = XINE_SPEED_FAST_4;
			break;
		default:
			return;
	}
	
	gtk_xine_set_param(mwin->gtx, XINE_PARAM_SPEED, new_speed);
	if(cur_speed == XINE_SPEED_PAUSE)
		xfmedia_mainwin_set_pause_buttons(mwin);
}


KEYBIND_FUNC(kb_slower)
{
	gint cur_speed = gtk_xine_get_param(mwin->gtx, XINE_PARAM_SPEED);
	gint new_speed;
	
	switch(cur_speed) {
		case XINE_SPEED_SLOW_4:
			new_speed = XINE_SPEED_PAUSE;
			break;
		case XINE_SPEED_SLOW_2:
			new_speed = XINE_SPEED_SLOW_4;
			break;
		case XINE_SPEED_NORMAL:
			new_speed = XINE_SPEED_SLOW_2;
			break;
		case XINE_SPEED_FAST_2:
			new_speed = XINE_SPEED_NORMAL;
			break;
		case XINE_SPEED_FAST_4:
			new_speed = XINE_SPEED_FAST_2;
			break;
		default:
			return;
	}
	
	gtk_xine_set_param(mwin->gtx, XINE_PARAM_SPEED, new_speed);
	if(new_speed == XINE_SPEED_PAUSE)
		xfmedia_mainwin_set_play_buttons(mwin);
}

KEYBIND_FUNC(kb_forward_10s)
{
	gint pos_stream, pos_time, length_time, new_time;
	
	if(gtk_xine_get_status(mwin->gtx) != XINE_STATUS_PLAY)
		return;
	
	if(gtk_xine_get_pos_length(mwin->gtx, &pos_stream, &pos_time, &length_time)) {
		new_time = pos_time + 10000;
		if(new_time > length_time)
			pos_time = length_time - 1000;
		gtk_xine_play(mwin->gtx, 0, new_time);
	}
}

KEYBIND_FUNC(kb_forward_30s)
{
	gint pos_stream, pos_time, length_time, new_time;
	
	if(gtk_xine_get_status(mwin->gtx) != XINE_STATUS_PLAY)
		return;
	
	if(gtk_xine_get_pos_length(mwin->gtx, &pos_stream, &pos_time, &length_time)) {
		new_time = pos_time + 30000;
		if(new_time > length_time)
			pos_time = length_time - 1000;
		gtk_xine_play(mwin->gtx, 0, new_time);
	}
}

KEYBIND_FUNC(kb_forward_60s)
{
	gint pos_stream, pos_time, length_time, new_time;
	
	if(gtk_xine_get_status(mwin->gtx) != XINE_STATUS_PLAY)
		return;
	
	if(gtk_xine_get_pos_length(mwin->gtx, &pos_stream, &pos_time, &length_time)) {
		new_time = pos_time + 60000;
		if(new_time > length_time)
			pos_time = length_time - 1000;
		gtk_xine_play(mwin->gtx, 0, new_time);
	}
}

KEYBIND_FUNC(kb_backward_10s)
{
	gint pos_stream, pos_time, length_time, new_time;
	
	if(gtk_xine_get_status(mwin->gtx) != XINE_STATUS_PLAY)
		return;
	
	if(gtk_xine_get_pos_length(mwin->gtx, &pos_stream, &pos_time, &length_time)) {
		new_time = pos_time - 10000;
		if(new_time < 0)
			pos_time = 0;
		gtk_xine_play(mwin->gtx, 0, new_time);
	}
}

KEYBIND_FUNC(kb_backward_30s)
{
	gint pos_stream, pos_time, length_time, new_time;
	
	if(gtk_xine_get_status(mwin->gtx) != XINE_STATUS_PLAY)
		return;
	
	if(gtk_xine_get_pos_length(mwin->gtx, &pos_stream, &pos_time, &length_time)) {
		new_time = pos_time - 30000;
		if(new_time < 0)
			pos_time = 0;
		gtk_xine_play(mwin->gtx, 0, new_time);
	}
}

KEYBIND_FUNC(kb_backward_60s)
{
	gint pos_stream, pos_time, length_time, new_time;
	
	if(gtk_xine_get_status(mwin->gtx) != XINE_STATUS_PLAY)
		return;
	
	if(gtk_xine_get_pos_length(mwin->gtx, &pos_stream, &pos_time, &length_time)) {
		new_time = pos_time - 60000;
		if(new_time < 0)
			pos_time = 0;
		gtk_xine_play(mwin->gtx, 0, new_time);
	}
}

KEYBIND_FUNC(kb_add_file)
{
	xfmedia_playlist_add_file_cb(mwin->playpause_btn, mwin);
}

KEYBIND_FUNC(kb_add_url)
{
	xfmedia_playlist_add_url_cb(mwin->playpause_btn, mwin);
}

KEYBIND_FUNC(kb_open_playlist)
{
	xfmedia_playlist_open_cb(mwin->playpause_btn, mwin);
}

KEYBIND_FUNC(kb_save_playlist)
{
	xfmedia_playlist_save_as_cb(mwin->playpause_btn, mwin);
}

KEYBIND_FUNC(kb_clear_playlist)
{
	xfmedia_playlist_new_cb(mwin->playpause_btn, mwin);
}

KEYBIND_FUNC(kb_av_sync_plus)
{
	gint cur_sync;
	
	cur_sync = gtk_xine_get_param(mwin->gtx, XINE_PARAM_AV_OFFSET);
	gtk_xine_set_param(mwin->gtx, XINE_PARAM_AV_OFFSET, cur_sync+3600);
}

KEYBIND_FUNC(kb_av_sync_minus)
{
	gint cur_sync;
	
	cur_sync = gtk_xine_get_param(mwin->gtx, XINE_PARAM_AV_OFFSET);
	gtk_xine_set_param(mwin->gtx, XINE_PARAM_AV_OFFSET, cur_sync-3600);
}

static XfMediaKeybindEntry keybind_entries[] =
{
	{ "play", kb_play, NULL },
	{ "pause", kb_pause, NULL },
	{ "stop", kb_stop, NULL },
	{ "prev-track", kb_prev_track, NULL },
	{ "next-track", kb_next_track, NULL },
	{ "quit", kb_quit, NULL },
	{ "toggle-fullscreen", kb_fullscreen, NULL },
	{ "toggle-shuffle", kb_shuffle, NULL },
	{ "toggle-repeat", kb_repeat, NULL },
	{ "toggle-audio-vis", kb_audio_vis, NULL },
	{ "speed-faster", kb_faster, NULL },
	{ "speed-slower", kb_slower, NULL },
	{ "forward-10s", kb_forward_10s, NULL },
	{ "forward-30s", kb_forward_30s, NULL },
	{ "forward-60s", kb_forward_60s, NULL },
	{ "backward-10s", kb_backward_10s, NULL },
	{ "backward-30s", kb_backward_30s, NULL },
	{ "backward-60s", kb_backward_60s, NULL },
#if 0  /* unimplemented (yet) */
	{ "toggle-enqueue-item", kb_enqueue, NULL },
	{ "toggle-mute", kb_mute, NULL },
#endif
	{ "add-file", kb_add_file, NULL },
	{ "add-url", kb_add_url, NULL },
	{ "open-playlist", kb_open_playlist, NULL },
	{ "save-playlist", kb_save_playlist, NULL },
	{ "clear-playlist", kb_clear_playlist, NULL },
	{ "av-sync-plus", kb_av_sync_plus, NULL },
	{ "av-sync-minus", kb_av_sync_minus, NULL },
	{ NULL, NULL, NULL }
};

void
xfmedia_keybindings_activate(XfMediaMainwin *mwin, const gchar *keybind_id)
{
	/* FIXME: put all this stuff in a GHashTable? */
	gint i;
	
	for(i = 0; keybind_entries[i].keybind_id; i++) {
		if(!g_ascii_strcasecmp(keybind_id, keybind_entries[i].keybind_id)) {
			keybind_entries[i].callback(mwin, keybind_entries[i].user_data);
			break;
		}
	}
}

enum {
	KB_LIST_FUNCTION = 0,
	KB_LIST_KEYBIND,
	KB_N_COLS
};

GtkWidget *
xfmedia_keybindings_create_editor(XfMediaMainwin *mwin, GtkWidget *parent)
{
	GtkWidget *topvbox, *sw, *treeview;
	GtkListStore *ls;
	GList *keybindings, *l;
	GtkTreeIter itr;
	GtkTreeViewColumn *col;
	GtkCellRenderer *render;
	
	
	topvbox = gtk_vbox_new(FALSE, BORDER);
	gtk_container_set_border_width(GTK_CONTAINER(topvbox), BORDER/2);
	gtk_widget_show(topvbox);
	
	sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
			GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	gtk_widget_show(sw);
	gtk_box_pack_start(GTK_BOX(topvbox), sw, TRUE, TRUE, 0);
	
	ls = gtk_list_store_new(KB_N_COLS, G_TYPE_STRING, G_TYPE_STRING);
	keybindings = xfmedia_keybindings_get_list();
	for(l = keybindings; l; l = l->next) {
		XfMediaKeybinding *kb = l->data;
		gchar *kb_str;
		
		kb_str = g_strdup_printf("%s%s%s%s%s",
				(kb->modmask & GDK_CONTROL_MASK) ? "ctrl+" : "",
				(kb->modmask & GDK_MOD1_MASK) ? "alt+" : "",
				(kb->modmask & GDK_MOD4_MASK) ? "super+" : "",
				(kb->modmask & GDK_SHIFT_MASK) ? "shift+" : "",
				gdk_keyval_name(kb->keyval));
		
		gtk_list_store_append(ls, &itr);
		gtk_list_store_set(ls, &itr, KB_LIST_FUNCTION, kb->keybind_id,
				KB_LIST_KEYBIND, kb_str, -1);
		g_free(kb_str);
	}
	xfmedia_keybindings_free_list(keybindings);
	
	treeview = gtk_tree_view_new_with_model(GTK_TREE_MODEL(ls));
	gtk_widget_set_sensitive(treeview, FALSE);
	gtk_widget_show(treeview);
	gtk_container_add(GTK_CONTAINER(sw), treeview);
	
	render = gtk_cell_renderer_text_new();
	col = gtk_tree_view_column_new_with_attributes(_("Function"), render,
			"text", KB_LIST_FUNCTION, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), col);
	
	render =gtk_cell_renderer_text_new();
	col = gtk_tree_view_column_new_with_attributes(_("Keys"), render,
			"text", KB_LIST_KEYBIND, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), col);
	
	return topvbox;
}
